import numpy as np
import ast
from tifffile import TiffFile
import matplotlib.pyplot as plt
from copy import deepcopy
from os.path import join

geneNms = ['AG',
           'AHP6',
           'ANT',
           'AP1',
           'AP2',
           'AP3',
           'AS1',
           'ATML1',
           'CLV3',
           'CUC1_2_3',
           'ETTIN',
           'FIL',
           'LFY',
           'MP',
           'PHB_PHV',
           'PI',
           'PUCHI',
           'REV',
           'SEP1',
           'SEP2',
           'SEP3',
           'STM',
           'SUP',
           'SVP',
           'WUS']

class Vec3(object):
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

    def dist(self, v1):
        return np.sqrt((self.x - v1.x)**2 +
                       (self.y - v1.y)**2 +
                       (self.z - v1.z)**2)

    def toOrganism(self, sep=' '):
        return sep.join([str(self.x), str(self.y), str(self.z)])

    def swapZX(self):
        xt = self.x
        self.x = self.z
        self.z = xt

    def toArray(self):
        return np.array([self.x, self.y, self.z])

    def __repr__(self):
        return self.toOrganism()

class Cell(object):
    def __init__(self, cid, pos, vol, exprs=None):
        self.cid = cid        #Int
        self.pos = pos        #Vec3
        self.vol = vol        #Real
        self.exprs = exprs    #Map String Bool

    def dist(self, c1):
        return self.pos.dist(c1.pos)

    def addExprs(self, exprs, geneNms):
        for gn in geneNms:
            if not gn in exprs.keys():
                exprs[gn] = False

        self.exprs = exprs
        self.geneNms = geneNms
        
        return

    def updateExprs(self, gfs):
        gexprsUpd = dict()
        
        for g, _ in self.exprs.items():
            if g in gfs:
                gf = gfs[g]
                (acts, reprs, out, fsAct, fsRepr) = gf
                inputs = acts + reprs
                vals = [self.exprs[i] for i in inputs]
                gexprsUpd[g] = evalI((acts, reprs, out), vals, fsAct, fsRepr)
            else:
                gexprsUpd[g] = self.exprs[g]
            
        ncell = Cell(self.cid, self.pos, self.vol)
        ncell.addExprs(gexprsUpd, deepcopy(self.geneNms))

        return ncell

    def updateExprsC(self, c):
        exprsUpd = dict([(g, c.exprs[g]) for g, v in self.exprs.items()])
        nc = Cell(self.cid, self.pos, self.vol, exprsUpd)
        nc.geneNms = self.geneNms
        return nc

    def _boolToInt(self, b):
        if b: return 1
 
        return 0
    
    def _exprsToOrg(self, sep=' '):
        exprs = [str(self._boolToInt(self.exprs[gn])) for gn in self.geneNms]
        
        return sep.join(exprs)

    def swapZX(self):
        self.pos.swapZX()

    @classmethod
    def fromTV(cls, nelems):
        def parsePos(spos):
            pos = list(np.fromstring(spos.replace('[', '').replace(']', ''),
                                     sep=' '))
            return pos
        
        cid = int(nelems['cid'])
        vol = float(nelems['volume'])
        [xc, yc, zc] = parsePos(nelems['center'])
        pos = Vec3(xc, yc, zc)

        return cls(cid, pos, vol)

    @classmethod
    def fromImage(cls, cd, res):
        volVoxel = np.prod(res)
        xc, yc, zc = cd.centroid
        nVoxels = cd.area

        xc, yc, zc = np.multiply(np.array([xc, yc, zc]), res)
        
        vol = nVoxels * volVoxel
        pos = Vec3(xc, yc, zc)
        cid = cd.label

        return cls(cid, pos, vol)

    def toOrganism(self):
        if self.exprs:
            return (' '.join([self.pos.toOrganism(), str(self.vol),
                          str(self.cid), self._exprsToOrg()]))
        else:
            return (' '.join([self.pos.toOrganism(), str(self.vol),
                              str(self.cid)]))

    def toText(self, sep=" "):
        if self.exprs:
            return (sep.join([str(self.cid),
                              self.pos.toOrganism(sep=sep),
                              str(self.vol),
                              self._exprsToOrg(sep=sep)]))
        else:
            return sep.join([str(self.cid), self.pos.toOrganism(sep=sep)])

    def centreToArray(self):
        return self.pos.toArray()

    def getOnGenes(self):
        return frozenset([g for g, st in self.exprs.items() if st]) - frozenset(['WUS'])

    def __repr__(self):
        return self.toOrganism()

class Tissue(object):
    def __init__(self, cells, neighs, ecellIds = None):
        self.cells = cells            #Map Int Cell
        self.neighs = neighs          #Map Int (Map Int Dist)
        self.L1_cells = ecellIds      #Set Int

    def __iter__(self):
        #iterate over a list instead of map
        return iter([v for _, v  in self.cells.items()])

    @classmethod
    def _extractCells(cls, data, res):
        from skimage.measure import regionprops
        cellsSeg = regionprops(data)

        cells = dict()
        for cd in cellsSeg:
            cells[int(cd.label)] = Cell.fromImage(cd, res)

        return cells

    @classmethod
    def _getConn(cls, cells, data):
        from skimage.future.graph import RAG

        connG = RAG(data)
        neigh = dict()
        
        for n in connG.nodes:
            nbrs = connG.adj[n].keys()
            ds = [cells[n].dist(cells[nb]) for nb in nbrs]
            
            neigh[n] = dict(zip(nbrs, ds))

        return neigh

    @classmethod
    def fromTV(cls, fnGeom, fnNeigh):
        def parseLn(ln):
            ln = str.strip(ln)
            nelems = dict()
            elems = str.split(ln, ',')
            for elm in elems:
                [name, val] = str.split(elm, ':')
                name = str.strip(name)
                val = str.strip(val)
                nelems[name] = val
                
            return nelems
        
        cells = dict() 
        neighs = dict()
        ecellIds = []
        
        with open(fnGeom, 'r') as f:
            for ln in f:
                nelems = parseLn(ln)
                c = Cell.fromTV(nelems)
                cells[c.cid] = c

        with open(fnNeigh, 'r') as f:
            for ln in f:
                nelems = ast.literal_eval(ln)
                cid = nelems['cid']
                ns = nelems['neighbors ids']

                if 1 in ns:
                    ecellIds.append(cid)
                    ns.remove(1)
                    
                neighs[cid] = dict((n, cells[cid].dist(cells[n])) for n in ns)

        return cls(cells, neighs, set(ecellIds))
    
    @classmethod
    def fromImage(cls, data, res):
        cells = cls._extractCells(data, res)
        conn = cls._getConn(cells, data)

        return cls(cells, conn)

    def adjView(self):
        return dict([(cid, ds.keys()) for cid, ds in self.neighs.items()])

    def relabelFromMap(self, relabelM):
        neighsR = []
        
        for c in self:
            c.cid = relabelM[c.cid]

        for c, dists in self.neighs.items():
            reNs = dict([(relabelM[n], d) for n, d in dists.items()])
            neighsR.append((relabelM[c], reNs))

        self.neighs = dict(neighsR)

        return
            
    def relabel(self):
        relabelM = dict([(c.cid, i) for i, c in enumerate(self)])

        self.relabelFromMap(relabelM)

        return

    def filterL1(self):
        if not self.neighs:
            return
        
        L1_cell_ids = self.neighs[1]

        L1_cells = dict([(c.cid, c) for c in self if c.cid in L1_cell_ids])
        L1_neighs = dict([(cid, nbrs) for cid, nbrs in deepcopy(self.neighs.items())
                           if cid in L1_cells_ids])

        return Tissue(L1_cells, L1_neighs)

    def toNeigh(self):
        
        def cellToNeigh(cid, nbs):
            nbsIds = nbs.keys()
            nNbs = len(nbsIds)
            nbsDists = [nbs[nid] for nid in nbsIds]
            
            return " ".join([str(cid), str(nNbs)]
                           + map(str, nbsIds)
                           + map(str, nbsDists))

        ncells = sum(1 for _ in self)
        header = str(ncells) + " " + "1"
        
        cellReprs = ([cellToNeigh(cid, nbs)
                      for cid, nbs in self.neighs.items()])

        return "\n".join([header] + cellReprs)
        
    def toOrganism(self):
        headerComment = "#x y z vol id"
        ns = str(len(self.cells)) + " 5"
        content = '\n'.join([cell.toOrganism() for cell in self])

        return "\n".join([headerComment, ns, content])

    def centresToArray(self):
        return np.array([c.centreToArray() for c in self])

    
class STissue(Tissue):
    def __init__(self, cells, neighs, gexprs=None, geneNms=None):
        super(STissue, self).__init__(cells, neighs)
        if gexprs and geneNms:
            self.addExprs(gexprs, geneNms)

    def _defGeneExprs(self):
        return dict([(gn, False) for gn in self.geneNms])
    
    def addExprs(self, gexprs, geneNms):
        self.geneNms = geneNms
        for cell in self:
            try:
                cell.addExprs(gexprs[cell.cid], geneNms)
            except KeyError:
                cell.addExprs(self._defGeneExprs(), geneNms)
                
        return

    def relabel(self):
        super(STissue, self).relabel()

    @classmethod
    def fromImage(cls, data, res, gexprs, geneNms):
        ts = Tissue.fromImage(data, res)
        return cls(ts.cells, ts.neighs, gexprs, geneNms)

    @classmethod
    def fromTV(cls, fnGeom, fnNeigh, gexprs, geneNms):
        ts = Tissue.fromTV(fnGeom, fnNeigh)
        return cls(ts.cells, ts.neighs, gexprs, geneNms)

    def toOrganism(self):
        headerComment = "#x y z vol id " + ' '.join(self.geneNms)
        ns = str(len(self.cells)) + " " + str(len(self.geneNms)+5)
        content = '\n'.join([cell.toOrganism() for cell in self])

        return "\n".join([headerComment, ns, content])

    def updateExprsBF(self, gfs):
        ncells = dict([(c.cid, c.updateExprs(gfs)) for c in self])
        return STissue(ncells, deepcopy(self.neighs))

    def updateExprsTs(self, ts, iLn):
        ncells = {}
        for c1 in self:
            try:
                c = ts.cells[iLn[c1.cid]]
            except KeyError:
                c = Cell(0, Vec3(0, 0, 0), 0, dict([(g, False) for g in geneNms]))

            ncells[c1.cid] = c1.updateExprsC(c)
                
        return STissue(ncells, deepcopy(self.neighs))

    def _boolToInt(b):
        if b: return 1
        else: return 0

    def toDict(self):
        tsDict = {}

        for c in self:
            tsDict[c.cid] = dict(((g, self._boolToInt(v)) for g, v in c.exprs))
        
        return tsDict

    def toCSV(self):
        headerComment = "id,x,y,z,vol," + ','.join(self.geneNms)

        content = '\n'.join([cell.toText(sep=",") for cell in self])

        return "\n".join([headerComment, content])


############### reading
def getResolution(f):
    try:
        x1, x2 = f.pages.pages[0].tags['XResolution']
        y1, y2 = f.pages.pages[0].tags['YResolution']
        xres = x2 / x1
        yres = y2 / y1
        zres = 1.0
    except:
        xres = 1.0
        yres = 1.0
        zres = 1.0

    return np.array([xres, yres, zres])

def readImage(fseg):
    f = TiffFile(fseg)
    seg = f.asarray().astype(int)

    return seg

def readImageRegionPoints(fseg, res=[1, 1, 1], d=2):
    from skimage.measure import regionprops
    
    cell_regions = dict()
    
    seg = readImage(fseg)
    cellsSeg = regionprops(seg)

    for cregion in cellsSeg:
        cell_regions[cregion.label] = np.multiply(cregion.coords[:, :d], res[:d])

    return cell_regions

def readRegionVolume(fseg, res=[1, 1, 1]):
    from skimage.measure import regionprops
    
    cell_regions = dict()
    
    seg = readImage(fseg)
    cellsSeg = regionprops(seg)

    for cregion in cellsSeg:
        cell_regions[cregion.label] = cregion.area*np.product(res)

    return cell_regions
    

def mergePoints(cell_regions, cids):
    return np.vstack(tuple([cell_regions[cid] for cid in cids]))

def getCellRegions(dataDir, fid="01", d=2):
    segFn = "yr{fid}/segmentations/YR_{fid}_{t}h_segmented.tif"
    ressFn = "yr" + fid + "/resolutions.txt"
    
    ress = readRess(join(dataDir, ressFn))
    tpoints = sorted(ress.keys())

    fsegs = dict([(t, join(dataDir, segFn.format(fid=fid, t=str(t)))) for t in tpoints])

    return dict([(t, readImageRegionPoints(fn, res=ress[t], d=d))
                   for t, fn in fsegs.items()])

def getCellRegionVolumesF1(dataDir, ress):
    fsegs = {10: "yr01/segmentations/YR_01_10h_segmented.tif",
             40: "yr01/segmentations/YR_01_40h_segmented.tif",
             96: "yr01/segmentations/YR_01_96h_segmented.tif",
             120: "yr01/segmentations/YR_01_120h_segmented.tif",
             132: "yr01/segmentations/YR_01_132h_segmented.tif"}

    return dict([(t, readRegionVolume(join(dataDir, fn), res=ress[t]))
                   for t, fn in fsegs.items()])

def readExprs(fn):
    #returns: Map Int (Map String Bool)  Cid -> (GeneName -> {On/Off})
    gexprs = dict() 
    
    with open(fn) as f:
        geneNms = f.readline().rstrip().split(' ')[1:]

        for line in f:
            els = line.rstrip().split(' ')
            cid = els[0]
            exprs = els[1:]
            gexprs[int(cid)] = dict(zip(geneNms, [bool(int(exp))
                                                  for exp in exprs]))
        
    return gexprs

def writeOrg(fnGeom, fnNeigh, ts):
    with open(fnGeom, 'w+') as f:
        f.write(ts.toOrganism())

    with open(fnNeigh, 'w+') as f:
        f.write(ts.toNeigh())

    return

def readRess(fn):
    import ast
    
    with open(fn, 'r') as f:
        sress = ("{" + f.read().replace("=", ":").replace("h", "").
                 replace("\n", "").replace("t", "") + "}")
        ress = ast.literal_eval(sress)

    return ress

#functions for common tasks
def mkOrgs(d, dExprs, dOut):
    from os.path import join
    
    timepoints = [10, 40, 96, 120, 132]
    geomF = "_segmented_tvformat_0_volume_position.txt"
    neighF = "_segmented_tvformat_0_neighbors_py.txt"
    exprF = "h.txt"

    initF = ".init"
    orgNeighF = ".neigh"

    for t in timepoints:
        print(t)
        fnGeom = join(d, "t" + str(t) + geomF)
        fnNeigh = join(d, "t" + str(t) + neighF)
        fnExpr = join(dExprs, "t_" + str(t) + exprF)
        fnGeomOut = join(dOut, "t" + str(t) + initF)
        fnNeighOut = join(dOut, "t" + str(t) + orgNeighF)

        gexprs = readExprs(fnExpr)
        ts = STissue.fromTV(fnGeom, fnNeigh, gexprs, geneNms)

        writeOrg(fnGeomOut, fnNeighOut, ts)

    return

def mkOrgGeomState(fim, fseg, fexpr, fout):
    (data, res) = readImage(fim, fseg)
    gexprs = readExprs(fexpr)

    ts = STissue.fromImage(data, res, gexprs, geneNms)

    writeOrg(fout, ts)

def mkOrgGeom(fim, fseg, fout):
    (data, res) = readImage(fim, fseg)
    ts = Tissue.fromImage(data, res)

    writeOrg(fout, ts)


def plotTissue(ts, hl):
    xs = [c.pos.x for c in ts]
    ys = [c.pos.y for c in ts]
    zs = [c.pos.z for c in ts]

    cs = ['red' if c.cid in hl else 'blue' for c in ts]

    fig = plt.figure()                                          
    ax = fig.add_subplot(211, projection='3d')
    ax.view_init(elev=95, azim=270)
    ax.scatter(xs, ys, zs, c=cs, alpha=0.9, s=200,
               edgecolors='black', linewidths=0.5)
    
    plt.show()

def filterL1(ts):
    if not ts.neighs:
        return
        
    L1_cell_ids = ts.neighs[1]

    L1_cells = dict([(c.cid, c) for c in ts if c.cid in L1_cell_ids])
    L1_neighs = dict([(cid, nbrs) for cid, nbrs in deepcopy(list(ts.neighs.items()))
                       if cid in L1_cell_ids])

    return Tissue(L1_cells, L1_neighs)

def const2(a, b): return b

def folds1(vals, fs, iVal):
    acc = iVal
    for i, v in enumerate(vals):
        acc = fs[i](acc, v)
        
    return acc

def folds(vals, fs):
    fs = (const2, ) + fs
    
    return folds1(vals, fs, False)

def evalI(intr, vals, fsAct, fsRepr):
    acts = intr[0]
    reprs = intr[1]
    
    nActs = len(acts)
    nReprs = len(reprs)
    
    if vals[nActs:]:
        return (folds(vals[:nActs], fsAct) and
                not folds(vals[nActs:], fsRepr))
    else:
        return folds(vals[:nActs], fsAct)
